package com.sbh.background_locations

import android.Manifest
import android.annotation.TargetApi
import android.app.Activity
import android.content.*
import android.content.pm.PackageManager
import android.location.Location
import android.os.Build
import android.os.IBinder
import android.text.format.DateFormat
import android.util.Log
import android.widget.Toast
import androidx.core.app.ActivityCompat
import androidx.localbroadcastmanager.content.LocalBroadcastManager
import com.karumi.dexter.Dexter
import com.karumi.dexter.MultiplePermissionsReport
import com.karumi.dexter.PermissionToken
import com.karumi.dexter.listener.PermissionRequest
import com.karumi.dexter.listener.multi.MultiplePermissionsListener
import com.sbh.background_locations.adapters.SBHConfig
import com.sbh.background_locations.callbacks.MotionCallback
import com.sbh.background_locations.callbacks.SBHCallback
import com.sbh.background_locations.callbacks.SBHLocationCallback
import com.sbh.background_locations.callbacks.ServiceCallback
import com.sbh.background_locations.evets.LocationProviderChangeEvent
import com.sbh.background_locations.models.SBHLocation
import com.sbh.background_locations.services.FirebaseDatabaseService
import com.sbh.background_locations.utils.SharedPreferenceUtil
import com.sbh.background_locations.utils.SingletonHolder
import java.util.*
import kotlin.collections.HashMap

class BackgroundLocation() {
    private var myReceiver: MyReceiver? = null
    private var mService: LocationUpdatesService? = null
    private var mBound: Boolean = false
    private var mActivity: Activity? = null
    private var mLocationCallback: SBHLocationCallback? = null
    private var mMotionCallback: MotionCallback? = null
    private var mOdometer: Double = 0.0
    private val mRunningQOrLater: Boolean = Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q
    private var mLastLocations: MutableList<SBHLocation> = mutableListOf()
    private var mIsMoving: Boolean = false
    private var mWaitingStart: Calendar? = null
    private var mServiceCallback: ServiceCallback? = null
    private var mStartCallback: SBHCallback? = null
    private var mReady = false
    private var mInitialized = false

    init {
        myReceiver = MyReceiver()
    }

    companion object : SingletonHolder<BackgroundLocation, Activity?>(::BackgroundLocation) {
        const val TAG = "BackgroundLocation"
        const val EVENT_LOCATION = "location"
        const val EVENT_MOTION_CHANGE = "motionchange"
        const val REQUEST_PERMISSIONS_REQUEST_CODE = 34
        const val ACTION_GET_PROVIDER_STATE = "getProviderState"
    }

    constructor(activity: Activity?) : this() {
        mActivity = activity
    }

    private val mServiceConnection = object : ServiceConnection {
        override fun onServiceConnected(name: ComponentName, service: IBinder) {
            Log.d(TAG, "SERVICE CONNECTED")
            mBound = true
            val binder = service as LocationUpdatesService.LocalBinder
            mService = binder.service
            mServiceCallback?.onConnected()
        }

        override fun onServiceDisconnected(name: ComponentName) {
            Log.d(TAG, "SERVICE DISCONNECTED")
            mService = null
            mServiceCallback?.onDisconnected()
        }
    }

    private fun requestLocation() {
        Log.d(TAG, "REQUESTING LOCATION UPDATES")
        if (!checkPermissions()) {
            requestPermissions()
        } else {
            if (mService == null) {
                bindService(object : ServiceCallback {
                    override fun onConnected() {
                        mServiceCallback = null
                        requestLocation()
                    }

                    override fun onDisconnected() {
                        mServiceCallback = null
                    }
                })
            } else {
                FirebaseDatabaseService.getInstance(mActivity?.applicationContext!!).initialize(mActivity?.applicationContext!!)
                Log.d(TAG, "requestLocation => Service => ${mService != null}")
                mService!!.startTrackingLocation()
                mStartCallback?.onSuccess()
                mStartCallback = null
            }
        }
    }

    internal fun onRequestPermissionsResult(requestCode: Int, permissions: Array<out String>?, grantResults: IntArray?): Boolean {
        Log.d(TAG, "onRequestPermissionsResult => Service => ${mService != null}")
        Log.i(TAG, "onRequestPermissionResult")
        if (requestCode == REQUEST_PERMISSIONS_REQUEST_CODE) {
            when {
                grantResults!!.isEmpty() -> Log.i(TAG, "User interaction was cancelled.")
                grantResults[0] == PackageManager.PERMISSION_GRANTED -> requestLocation()
                else -> Toast.makeText(mActivity, R.string.permission_denied_explanation, Toast.LENGTH_LONG).show()
            }
        }
        return true

    }

    private fun checkPermissions(): Boolean {
        Log.d(TAG, " checkPermissions() => mRunningQOrLater => $mRunningQOrLater")
        return if (mRunningQOrLater) {
            (PackageManager.PERMISSION_GRANTED == ActivityCompat.checkSelfPermission(mActivity!!.applicationContext, Manifest.permission.ACCESS_FINE_LOCATION)
                    && PackageManager.PERMISSION_GRANTED == ActivityCompat.checkSelfPermission(mActivity!!.applicationContext, Manifest.permission.ACCESS_BACKGROUND_LOCATION))
        } else {
            PackageManager.PERMISSION_GRANTED == ActivityCompat.checkSelfPermission(mActivity!!.applicationContext, Manifest.permission.ACCESS_FINE_LOCATION)
        }
    }

    @TargetApi(Build.VERSION_CODES.P)
    private fun requestPermissions() {
        val shouldProvideRationale = mActivity.let {
            ActivityCompat.shouldShowRequestPermissionRationale(it!!, Manifest.permission.ACCESS_FINE_LOCATION)
                    || (mRunningQOrLater && ActivityCompat.shouldShowRequestPermissionRationale(it, Manifest.permission.ACCESS_BACKGROUND_LOCATION))
        }
        if (shouldProvideRationale) {

            Log.i(TAG, "Displaying permission rationale to provide additional context.")
            Toast.makeText(mActivity, R.string.permission_rationale, Toast.LENGTH_LONG).show()

        } else {
            Log.i(TAG, "Requesting permission")
            if (mRunningQOrLater) {
                mActivity.let {
                    ActivityCompat.requestPermissions(it!!,
                            arrayOf(Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_BACKGROUND_LOCATION),
                            REQUEST_PERMISSIONS_REQUEST_CODE)
                }
            } else {
                mActivity.let {
                    ActivityCompat.requestPermissions(it!!,
                            arrayOf(Manifest.permission.ACCESS_FINE_LOCATION),
                            REQUEST_PERMISSIONS_REQUEST_CODE)
                }
            }
        }
    }

    private fun bindService(callback: ServiceCallback) {
        mServiceCallback = callback
        Log.d(TAG, "bindService() => mBound => $mBound")
        mActivity.let {
            LocalBroadcastManager.getInstance(it!!.applicationContext).registerReceiver(myReceiver!!,
                    IntentFilter(LocationUpdatesService.ACTION_FOREGROUND_ONLY_LOCATION_BROADCAST))
            if (!mBound) {
                Log.d(TAG, "SERVICE BOUNDED")
                it.bindService(Intent(it, LocationUpdatesService::class.java), mServiceConnection, Context.BIND_AUTO_CREATE)
            }
        }
    }

    fun start(callback: SBHCallback) {
        mStartCallback = callback
        Log.d(TAG, "START => ${mActivity == null}")

        if (mActivity!!.applicationContext.let { SharedPreferenceUtil.getLocationTrackingPref(it) }) {
            requestPermissions()
        } else {
            requestLocation()
        }
    }

    fun stop() {
        mService?.stopTrackingLocation()

        mActivity!!.applicationContext.let { LocalBroadcastManager.getInstance(it).unregisterReceiver(myReceiver!!) }

        if (mBound) {
            try {
                mActivity!!.unbindService(mServiceConnection)
            } catch (e: Exception) {
                Log.e(TAG, "stop() => ${e.message}")
            }
            mBound = false
        }
    }

    fun onLocation(locationCallback: SBHLocationCallback) {
        mLocationCallback = locationCallback
    }

    fun onMotionChange(motionCallback: MotionCallback) {
        mMotionCallback = motionCallback
    }

    fun onActivityDestroy() {
//        mLocationCallback = null
    }

    fun getOdometer(): Double {
        return mOdometer
    }

    fun setOdometer(odometer: Double, callback: SBHLocationCallback) {
        mOdometer = odometer
        if (mLastLocations.isNotEmpty())
            callback.onLocation(mLastLocations.last())
        else
            callback.onError(0)
    }

    fun getCurrentPosition(callback: SBHLocationCallback) {
        Log.d(TAG, "getCurrentPosition")

        Dexter.withActivity(mActivity)
                .withPermissions(
                        Manifest.permission.ACCESS_COARSE_LOCATION,
                        Manifest.permission.ACCESS_FINE_LOCATION)
                .withListener(object : MultiplePermissionsListener {
                    override fun onPermissionsChecked(report: MultiplePermissionsReport?) {
                        report.let {
                            if (report!!.areAllPermissionsGranted()) {
                                if (mService == null) {
                                    bindService(object : ServiceCallback {
                                        override fun onConnected() {
                                            mService?.getCurrentLocation(object : SBHLocationCallback {
                                                override fun onLocation(location: SBHLocation) {
                                                    callback.onLocation(location)
                                                }

                                                override fun onError(error: Int) {
                                                    callback.onError(error)
                                                }
                                            })
                                            mServiceCallback = null
                                        }

                                        override fun onDisconnected() {
                                            mServiceCallback = null
                                        }

                                    })
                                } else {
                                    mService?.getCurrentLocation(object : SBHLocationCallback {
                                        override fun onLocation(location: SBHLocation) {
                                            callback.onLocation(location)
                                        }

                                        override fun onError(error: Int) {
                                            callback.onError(error)
                                        }
                                    })
                                }
                            }
                        }
                    }

                    override fun onPermissionRationaleShouldBeShown(
                            permissions: MutableList<PermissionRequest>?,
                            token: PermissionToken?
                    ) {
                        // Remember to invoke this method when the custom rationale is closed
                        // or just by default if you don't want to use any custom rationale.
                        token!!.continuePermissionRequest()
                    }
                })
                .withErrorListener {
                    //                    toast(it.name)
                }
                .check()
    }

    fun ready(callback: SBHCallback) {
        mReady = false
        val config = SBHConfig.getInstance(mActivity?.applicationContext!!)
        if (this.mInitialized) {
            if (config.getEnabled()) {
//                val var5 = (((Builder(this.m).setPersist(false) as Builder).setSamples(1) as Builder).setDesiredAccuracy(100) as Builder).setMaximumAge(60000L).build()
//                this.i.getCurrentPosition(var5)
            }
            FirebaseDatabaseService.getInstance(mActivity?.applicationContext!!)
            callback.onSuccess()
        } else {
//            this.e()
            this.mInitialized = true
            val var3: Boolean = if (config.hasSchedule() && config.getSchedulerEnabled()) {
                this.startSchedule()
                false
            } else {
                config.setSchedulerEnabled(false)
                config.getEnabled()
            }

            if (var3) {
                if (config.isLocationTrackingMode()) {
                    this.start(callback)
                } else {
                    this.startGeofences(callback)
                }
            } else {
                FirebaseDatabaseService.getInstance(mActivity?.applicationContext!!)
                callback.onSuccess()
            }

//            if (config.getEnabled() && config.hasUrl() && config.getAutoSync() && this.g.isNetworkAvailable()) {
//                this.g.flush()
//            }

//            TSLog.logger.debug(Application.tsAFd("篳\ue072\ue067哉\uf6ec\ue4b2\ue699걟ﲛÞ頶뇻톦鳉㙾论鐖袡䴴㫜") + com.transistorsoft.locationmanager.util.b.a(this.m))
//            val var4 = LocationProviderChangeEvent(this.m)
//            EventBus.getDefault().post(var4)
        }
    }

    fun setConfig(callback: SBHCallback) {
        FirebaseDatabaseService.getInstance(mActivity?.applicationContext!!).initialize(mActivity?.applicationContext!!)
        callback.onSuccess()
    }

    private fun incrementOdometer(distance: Float) {
        mOdometer = mOdometer.plus(distance)
    }

    fun getGpsState(): Boolean {
        return mService!!.checkGps()
    }

    fun getProviderState(): LocationProviderChangeEvent {
        return LocationProviderChangeEvent(mActivity!!.applicationContext)
    }

    private fun startSchedule() {

    }

    private fun startGeofences(callback: SBHCallback) {

    }

    private inner class MyReceiver : BroadcastReceiver() {
        private var lastLocation: Location? = null
        override fun onReceive(context: Context, intent: Intent) {
            val location = intent.getParcelableExtra<Location>(LocationUpdatesService.EXTRA_LOCATION)
            if (location != null) {
                val locationMap = HashMap<String, Any>()
                locationMap["timestamp"] = DateFormat.format("yyyy-MM-dd HH:mm:ss", location.time)

                if (lastLocation != null) {
                    val distance = lastLocation!!.distanceTo(location)
                    locationMap["odometer"] = distance
                    locationMap["isMoving"] = distance > 3

                    if (locationMap["isMoving"] as Boolean)
                        incrementOdometer(distance)
                }

                lastLocation = location
                val sbhLocation = SBHLocation(location = location, locationMap = locationMap)
                mLocationCallback?.onLocation(sbhLocation)

                addLocation(sbhLocation)
//                checkMovement()
            }
        }

        private fun addLocation(sbhLocation: SBHLocation) {
            if (mLastLocations.size >= 10) {
                mLastLocations.removeAt(0)
            }

            mLastLocations.add(sbhLocation)
            Log.d(TAG, "addLocation() => ${mLastLocations.size}")
        }

        private fun checkMovement() {
            mLastLocations.last().isMoving.let { mMotionCallback?.onMotionChange(it!!) }
//            TODO implement => distance, duration, accuracy
//            for (location in mLastLocations) {
//
//            }
//
//
//            val items = mLastLocations.filter {
//                it.isMoving != null && !it.isMoving!!
//            }
//
//            val notMoving = items.size >= 7
//
//            Log.d(TAG, "items => ${items.size}")
//            Log.d(TAG, "notMoving => $notMoving")
//            Log.d(TAG, "mIsMoving => $mIsMoving")
//
//            if(items.size >= 2 && mWaitingStart == null) {
//                Log.d(TAG, "TIMER STARTED")
//                mWaitingStart = Calendar.getInstance()
//            } else if(items.size < 3 && mWaitingStart != null) {
//                Log.d(TAG, "TIMER STOPPED => START MOVING")
//                mWaitingStart = null
//            }
//
//            if(items.size >= 8 && mIsMoving != notMoving) {
//                mIsMoving = notMoving
//                mMotionCallback?.onMotionChange(mIsMoving)
//            } else {
//                mIsMoving = notMoving
//                mMotionCallback?.onMotionChange(mIsMoving)
//            }
        }
    }
}