package com.sbh.background_locations

import android.annotation.SuppressLint
import android.app.Activity
import android.app.Application
import android.os.Bundle
import android.util.Log
import com.sbh.background_locations.adapters.SBHConfig
import com.sbh.background_locations.callbacks.SBHCallback
import com.sbh.background_locations.callbacks.SBHLocationCallback
import com.sbh.background_locations.models.SBHLocation
import com.sbh.background_locations.streams.LocationStreamHandler
import com.sbh.background_locations.streams.MotionChangeStreamHandler
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry
import io.flutter.plugin.common.PluginRegistry.Registrar
import io.flutter.view.FlutterNativeView
import org.json.JSONArray
import org.json.JSONException
import org.json.JSONObject
import java.util.*
import kotlin.collections.HashMap

class BackgroundLocationsPlugin() : MethodCallHandler, Application.ActivityLifecycleCallbacks, PluginRegistry.ViewDestroyListener {
    private lateinit var mActivity: Activity
    private lateinit var mMethodChannel: MethodChannel
    private lateinit var mRegistrar: Registrar
    private var mReady: Boolean = false

    companion object {
        private const val TAG: String = "BackgroundLocationsPlugin"
        const val PLUGIN_ID: String = "com.sbh/background_locations"
        const val METHOD_CHANNEL_NAME: String = "$PLUGIN_ID/methods"

        @JvmStatic
        fun registerWith(registrar: Registrar) {
            val methodChannel = MethodChannel(registrar.messenger(), METHOD_CHANNEL_NAME)
            val plugin = BackgroundLocationsPlugin(registrar, methodChannel)
            methodChannel.setMethodCallHandler(plugin)
            registrar.addViewDestroyListener(plugin)
        }

        private const val ACTION_RESET = "reset"
        private const val ACTION_READY = "ready"
        private const val ACTION_SET_CONFIG = "setConfig"

        private const val ACTION_REGISTER_HEADLESS_TASK = "registerHeadlessTask"
        private const val ACTION_GET_GPS_STATE = "getGpsState"
        private const val ACTION_GET_STATE = "getState"
        private const val ACTION_GET_LOG = "getLog"
        private const val ACTION_EMAIL_LOG = "emailLog"
        private const val ACTION_START_SCHEDULE = "startSchedule"
        private const val ACTION_STOP_SCHEDULE = "stopSchedule"
        private const val ACTION_LOG = "log"
        private const val ACTION_REQUEST_SETTINGS = "requestSettings"
        private const val ACTION_SHOW_SETTINGS = "showSettings"
        private const val ACTION_REGISTER_PLUGIN = "registerPlugin"
        private const val JOB_SERVICE_CLASS = "com.sbh.flutter.background_locations.HeadlessTask"
    }

    constructor(registrar: Registrar, channel: MethodChannel) : this() {
        this.mRegistrar = registrar
        this.mActivity = registrar.activity()
        this.mMethodChannel = channel

        this.mRegistrar.addRequestPermissionsResultListener { requestCode, permissions, grantResults ->
            BackgroundLocation.getInstance(mActivity).onRequestPermissionsResult(requestCode, permissions, grantResults)
        }

        LocationStreamHandler().register(registrar)
        MotionChangeStreamHandler().register(registrar)
    }

    @SuppressLint("LongLogTag")
    override fun onMethodCall(call: MethodCall, result: Result) {
        when {
            call.method == "getPlatformVersion" -> {
                result.success("Android " + android.os.Build.VERSION.RELEASE)
            }
            call.method.equals(ACTION_READY, ignoreCase = true) -> {
                val params = call.arguments as Map<String, Any>
                Log.d(TAG, params.toString())
                ready(params, result)
            }
            call.method.equals(ACTION_RESET, ignoreCase = true) -> {
                reset(call.arguments, result)
            }
            call.method.equals(ACTION_SET_CONFIG, ignoreCase = true) -> {
                val params = call.arguments as Map<String, Any>
                Log.d(TAG, params.toString())
                setConfig(params, result)
            }
            call.method == ACTION_GET_GPS_STATE -> {
                BackgroundLocation.getInstance(mActivity).getGpsState()
            }
            call.method == ACTION_GET_STATE -> {
                getState(result)
            }
            call.method == BackgroundLocation.ACTION_GET_PROVIDER_STATE -> {
                getProviderState(result)
            }
            call.method == "getCurrentPosition" -> {
                getCurrentPosition(result)
            }
            call.method == "getOdometer" -> {
                getOdometer(result)
            }
            call.method == "setOdometer" -> {
                setOdometer(call.arguments as Double, result)
            }
            call.method == "stop" -> {
                BackgroundLocation.getInstance(mActivity).stop()
            }
            call.method == "start" -> {
                start(result)
            }
            else -> result.notImplemented()
        }
    }

    private fun getState(result: Result) {
        resultWithState(result)
    }

    @SuppressLint("LongLogTag")
    private fun ready(params: Map<String, Any>, result: Result) {
        if (mReady) {
//            TSLog.logger.warn(TSLog.warn("#ready already called.  Redirecting to #setConfig"))
            setConfig(params, result)
            return
        }
//        val config = SBHConfig.getInstance(mActivity.applicationContext)
        if (!applyConfig(params, result)) {
            return
        }

//        if (config.isFirstBoot()) {
//            if (!applyConfig(params, result)) {
//                return
//            }
//        } else {
//            val reset = !params.containsKey("reset") || params["reset"] as Boolean
//            if (reset) {
//                config.reset()
//                if (!applyConfig(params, result)) {
//                    return
//                }
//            }
//        }
        mReady = true
        BackgroundLocation.getInstance(mActivity).ready(object : SBHCallback {
            override fun onSuccess() {
                resultWithState(result)
            }

            override fun onFailure(error: String) {
                result.error(error, null, null)
            }
        })
    }

    private fun getProviderState(result: Result) {
        result.success(BackgroundLocation.getInstance(mActivity).getProviderState().toMap())
    }

    private fun start(result: Result) {
        BackgroundLocation.getInstance(mActivity).start(object : SBHCallback {
            override fun onSuccess() {
                val map: MutableMap<String, Any> = HashMap()
                map["enabled"] = true
                result.success(map)
//                resultWithState(result)
            }

            override fun onFailure(error: String) {
                result.error(error, null, null)
            }
        })
    }

    private fun setConfig(config: Map<String, Any>, result: Result) {
        if (!applyConfig(config, result)) return
        BackgroundLocation.getInstance(mActivity).setConfig(object : SBHCallback {
            override fun onSuccess() {
                resultWithState(result)
            }

            override fun onFailure(error: String) {
                result.error(error, null, null)
            }
        })
    }

    private fun reset(args: Any?, result: Result) {
        val config: SBHConfig = SBHConfig.getInstance(mActivity)
        config.reset()

        if (args != null) {
            if (args.javaClass == java.util.HashMap::class.java) {
                val params: Map<String, Any> = args as HashMap<String, Any>
                if (!applyConfig(params, result)) return
            }
        }
        resultWithState(result)
    }

    private fun getOdometer(result: Result) {
        result.success(BackgroundLocation.getInstance(mActivity).getOdometer())
    }

    private fun setOdometer(odometer: Double, result: Result) {
        BackgroundLocation.getInstance(mActivity).setOdometer(odometer, object : SBHLocationCallback {
            override fun onLocation(location: SBHLocation) {
                result.success(location.toMap())
            }

            override fun onError(error: Int) {
                result.error(error.toString(), null, null)
            }
        })
    }

    private fun getCurrentPosition(result: Result) {
        BackgroundLocation.getInstance(mActivity).getCurrentPosition(object : SBHLocationCallback {
            override fun onLocation(location: SBHLocation) {
                result.success(location.toMap())
            }

            override fun onError(error: Int) {
                result.error(error.toString(), "unable to fetch location", null)
            }
        })
    }

    ////
    // Utility Methods
    //
    @Throws(JSONException::class)
    private fun mapToJson(map: Map<String, Any>): JSONObject {
        val jsonData = JSONObject()
        for (key in map.keys) {
            var value = map[key]
            if (value is Map<*, *>) {
                value = mapToJson((value as Map<String, Any>?)!!)
            } else if (value is List<*>) {
                value = listToJson((value as List<Any>?)!!)
            }
            jsonData.put(key, value)
        }
        return jsonData
    }

    @Throws(JSONException::class)
    private fun listToJson(list: List<Any>): JSONArray {
        val jsonData = JSONArray()
        for (value in list) {
            var _value = value
            if (_value is Map<*, *>) {
                _value = mapToJson((_value as Map<String, Any>?)!!)
            } else if (value is List<*>) {
                _value = listToJson((_value as List<Any>?)!!)
            }
            jsonData.put(_value)
        }
        return jsonData
    }

    @Throws(JSONException::class)
    private fun jsonToMap(json: Any): Map<String, Any> {
        var retMap: Map<String, Any> = HashMap()

        if (json !== JSONObject.NULL) {
            retMap = toMap(json as JSONObject)
        }
        return retMap
    }


    @Throws(JSONException::class)
    private fun toMap(`object`: JSONObject): Map<String, Any> {
        val map = HashMap<String, Any>()

        val keysItr = `object`.keys()
        while (keysItr.hasNext()) {
            val key = keysItr.next()
            var value = `object`.get(key)

            if (value is JSONArray) {
                value = toList(value)
            } else if (value is JSONObject) {
                value = toMap(value)
            }
            map[key] = value
        }
        return map
    }

    @Throws(JSONException::class)
    private fun toList(array: JSONArray): List<Any> {
        val list = ArrayList<Any>()
        for (i in 0 until array.length()) {
            var value = array.get(i)
            if (value is JSONArray) {
                value = toList(value)
            } else if (value is JSONObject) {
                value = toMap(value)
            }
            list.add(value)
        }
        return list
    }

    private fun setHeadlessJobService(config: MutableMap<String, Any>): Map<String, Any> {
        config["headlessJobService"] = JOB_SERVICE_CLASS
        return config
    }

    @SuppressLint("LongLogTag")
    private fun applyConfig(params: Map<String, Any>, result: Result): Boolean {
        Log.d(TAG, "applyConfig $params")
        val config = SBHConfig.getInstance(mActivity.applicationContext)

        try {
            config.updateWithJSONObject(mapToJson(setHeadlessJobService(params as MutableMap<String, Any>)))
        } catch (e: JSONException) {
            result.error(e.message, null, null)
            Log.e(TAG, e.message, e)
            return false
        }

        return true
    }

    private fun resultWithState(result: Result) {
        try {
            result.success(jsonToMap(SBHConfig.getInstance(mActivity.applicationContext).toJson()))
        } catch (e: JSONException) {
            e.printStackTrace()
            result.error(e.message, null, null)
        }

    }

    override fun onActivityPaused(activity: Activity?) {
        TODO("not implemented") //To change body of created functions use File | Settings | File Templates.
    }

    override fun onActivityResumed(activity: Activity?) {
        TODO("not implemented") //To change body of created functions use File | Settings | File Templates.
    }

    override fun onActivityStarted(activity: Activity?) {
        TODO("not implemented") //To change body of created functions use File | Settings | File Templates.
    }

    override fun onActivityDestroyed(activity: Activity?) {
        if (mRegistrar.activity() == null) {
            BackgroundLocation.getInstance(mActivity).onActivityDestroy()
            mActivity.application.unregisterActivityLifecycleCallbacks(this)
        }
    }

    override fun onActivitySaveInstanceState(activity: Activity?, outState: Bundle?) {
        TODO("not implemented") //To change body of created functions use File | Settings | File Templates.
    }

    override fun onActivityStopped(activity: Activity?) {
//        val config = SBHConfig.getInstance(activity)
//        if (config.getEnabled() && config.getEnableHeadless() && !config.getStopOnTerminate()) {
//            TSScheduleManager.getInstance(activity).oneShot(TerminateEvent.ACTION, 10000)
//        }
    }

    override fun onActivityCreated(activity: Activity?, savedInstanceState: Bundle?) {
        TODO("not implemented") //To change body of created functions use File | Settings | File Templates.
    }

    override fun onViewDestroy(p0: FlutterNativeView?): Boolean {
        return true
    }
}
