package com.sbh.background_locations.filters

import android.util.Log
import com.sbh.background_locations.common.Matrix

class GPSAccKalmanFilter() {
    val TAG = "GPSAccKalmanFilter"

    private var m_timeStampMsPredict: Long = 0
    private var m_timeStampMsUpdate: Long = 0
    private var m_predictCount: Int = 0
    private lateinit var m_kf: KalmanFilter
    private var m_accSigma: Double = 0.0
    private var m_useGpsSpeed: Boolean = false
    private var mVelFactor = 1.0
    private var mPosFactor = 1.0

    constructor (useGpsSpeed: Boolean,
                 x: Double, y: Double,
                 xVel: Double, yVel: Double,
                 accDev: Double, posDev: Double,
                 timeStampMs: Long,
                 velFactor: Double,
                 posFactor: Double) : this() {
        val mesDim = if (useGpsSpeed) 4 else 2
        m_useGpsSpeed = useGpsSpeed

        m_kf = KalmanFilter(4, mesDim, 2)
        m_timeStampMsUpdate = timeStampMs
        m_timeStampMsPredict = m_timeStampMsUpdate
        m_accSigma = accDev
        m_predictCount = 0
        m_kf.Xk_k.setData(x, y, xVel, yVel)

        m_kf.H.setIdentityDiag() //state has 4d and measurement has 4d too. so here is identity
        m_kf.Pk_k.setIdentity()
        m_kf.Pk_k.scale(posDev)
        mVelFactor = velFactor
        mPosFactor = posFactor
    }

    private fun rebuildF(dtPredict: Double) {
        m_kf.F.setData(1.0, 0.0, dtPredict, 0.0, 0.0, 1.0, 0.0, dtPredict, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0)
    }

    private fun rebuildU(xAcc: Double,
                         yAcc: Double) {
        m_kf.Uk.setData(xAcc, yAcc)
    }

    private fun rebuildB(dtPredict: Double) {
        val dt2 = 0.5 * dtPredict * dtPredict
        m_kf.B.setData(dt2, 0.0, 0.0, dt2, dtPredict, 0.0, 0.0, dtPredict)
    }

    private fun rebuildR(posSigma: Double, velSigma: Double) {
        var posSigma = posSigma
        var velSigma = velSigma

        posSigma *= mPosFactor
        velSigma *= mVelFactor

        Log.i(TAG, "rebuildR: { " +
                "velSigma : " + velSigma +
                ", posSigma : " + posSigma +
                ", velFactor : " + mVelFactor +
                ", posFactor :" + mPosFactor +
                "}")
        if (m_useGpsSpeed) {
            m_kf.R.setData(posSigma, 0.0, 0.0, 0.0, 0.0, posSigma, 0.0, 0.0, 0.0, 0.0, velSigma, 0.0, 0.0, 0.0, 0.0, velSigma)
        } else {
            m_kf.R.setIdentity()
            m_kf.R.scale(posSigma)
        }
    }

    private fun rebuildQ(dtUpdate: Double,
                         accDev: Double) {
        //        now we use predictCount. but maybe there is way to use dtUpdate.
        //        m_kf.Q.setIdentity();
        //        m_kf.Q.scale(accSigma * dtUpdate);
        val velDev = accDev * m_predictCount
        val posDev = velDev * m_predictCount / 2
        val covDev = velDev * posDev

        val posSig = posDev * posDev
        val velSig = velDev * velDev

        m_kf.Q.setData(posSig, 0.0, covDev, 0.0, 0.0, posSig, 0.0, covDev, covDev, 0.0, velSig, 0.0, 0.0, covDev, 0.0, velSig)
    }

    fun predict(timeNowMs: Long,
                xAcc: Double,
                yAcc: Double) {
        val dtPredict = (timeNowMs - m_timeStampMsPredict) / 1000.0
        val dtUpdate = (timeNowMs - m_timeStampMsUpdate) / 1000.0
        rebuildF(dtPredict)
        rebuildB(dtPredict)
        rebuildU(xAcc, yAcc)

        ++m_predictCount
        rebuildQ(dtUpdate, m_accSigma)

        m_timeStampMsPredict = timeNowMs
        m_kf.predict()
        Matrix.matrixCopy(m_kf.Xk_km1, m_kf.Xk_k)
    }

    fun update(timeStamp: Long,
               x: Double,
               y: Double,
               xVel: Double,
               yVel: Double,
               posDev: Double,
               velErr: Double) {
        m_predictCount = 0
        m_timeStampMsUpdate = timeStamp
        rebuildR(posDev, velErr)
        m_kf.Zk.setData(x, y, xVel, yVel)
        m_kf.update()
    }

    fun getCurrentX(): Double {
        return m_kf.Xk_k.data[0][0]
    }

    fun getCurrentY(): Double {
        return m_kf.Xk_k.data[1][0]
    }

    fun getCurrentXVel(): Double {
        return m_kf.Xk_k.data[2][0]
    }

    fun getCurrentYVel(): Double {
        return m_kf.Xk_k.data[3][0]
    }
}