/*
 * Copyright 2019 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.sbh.background_locations

import android.Manifest
import android.annotation.SuppressLint
import android.app.*
import android.content.Context
import android.content.Intent
import android.content.pm.PackageManager
import android.content.res.Configuration
import android.hardware.Sensor
import android.hardware.SensorEvent
import android.hardware.SensorEventListener
import android.hardware.SensorManager
import android.location.GnssStatus
import android.location.GpsStatus
import android.location.Location
import android.location.LocationManager
import android.os.*
import android.util.Log
import androidx.core.app.ActivityCompat.checkSelfPermission
import androidx.core.app.NotificationCompat
import androidx.localbroadcastmanager.content.LocalBroadcastManager
import com.google.android.gms.location.*
import com.sbh.background_locations.common.Coordinates
import com.sbh.background_locations.common.SensorGpsDataItem
import com.sbh.background_locations.common.Utils
import com.sbh.background_locations.filters.GPSAccKalmanFilter
import com.sbh.background_locations.filters.GeohashRTFilter
import com.sbh.background_locations.models.SBHLocation
import com.sbh.background_locations.services.FirebaseDatabaseService
import com.sbh.background_locations.utils.SharedPreferenceUtil
import java.util.concurrent.PriorityBlockingQueue
import java.util.concurrent.TimeUnit
import kotlin.math.sqrt

private const val TAG = "LocationUpdatesService"

private val UPDATE_INTERVAL_IN_MILLISECONDS: Long = TimeUnit.SECONDS.toMillis(2)

private val FASTEST_UPDATE_INTERVAL_IN_MILLISECONDS = UPDATE_INTERVAL_IN_MILLISECONDS / 2

private val MAX_WAIT_TIME: Long = TimeUnit.SECONDS.toMillis(2)

private const val NOTIFICATION_ID = 12345678

private const val NOTIFICATION_CHANNEL_ID = "flutter_background_location_channel_01"

class LocationUpdatesService : Service(), SensorEventListener {
    override fun onAccuracyChanged(sensor: Sensor?, accuracy: Int) {}

    override fun onSensorChanged(event: SensorEvent?) {
        if (event == null)
            return

        val east = 0
        val north = 1
        val up = 2

        val now = SystemClock.elapsedRealtimeNanos()
        val nowMs = Utils.nano2milli(now)
        when (event.sensor.type) {
            Sensor.TYPE_LINEAR_ACCELERATION -> {
                System.arraycopy(event.values, 0, linearAcceleration, 0, event.values.size)
                android.opengl.Matrix.multiplyMV(absAcceleration, 0, rotationMatrixInv,
                        0, linearAcceleration, 0)

//                val logStr = String.format("%d%d abs acc: %f %f %f",
//                        Utils.LogMessageType.ABS_ACC_DATA.ordinal,
//                        nowMs, absAcceleration[east], absAcceleration[north], absAcceleration[up])
//                log2File(logStr)

                if (m_kalmanFilter != null) {
                    val sdi = SensorGpsDataItem(nowMs,
                            SensorGpsDataItem.NOT_INITIALIZED.toDouble(),
                            SensorGpsDataItem.NOT_INITIALIZED.toDouble(),
                            SensorGpsDataItem.NOT_INITIALIZED.toDouble(),
                            absAcceleration[north],
                            absAcceleration[east],
                            absAcceleration[up],
                            SensorGpsDataItem.NOT_INITIALIZED,
                            SensorGpsDataItem.NOT_INITIALIZED,
                            SensorGpsDataItem.NOT_INITIALIZED,
                            SensorGpsDataItem.NOT_INITIALIZED.toDouble(),
                            m_magneticDeclination)
                    m_sensorDataQueue.add(sdi)
                }
            }
            Sensor.TYPE_ROTATION_VECTOR -> {
                SensorManager.getRotationMatrixFromVector(rotationMatrix, event.values)
                android.opengl.Matrix.invertM(rotationMatrixInv, 0, rotationMatrix, 0)
            }
        }
    }

    /*
     * Checks whether the bound activity has really gone away (foreground service with notification
     * created) or simply orientation change (no-op).
     */
    private var configurationChange = false

    private var serviceRunningInForeground = false

    private val localBinder = LocalBinder()

    private lateinit var firebaseDatabaseService: FirebaseDatabaseService
    private lateinit var notificationManager: NotificationManager

    private lateinit var fusedLocationProviderClient: FusedLocationProviderClient
    private lateinit var locationRequest: LocationRequest
    private lateinit var locationCallback: LocationCallback
    private lateinit var gnssStatusCallback: GnssStatus.Callback
    private lateinit var gpsStatusCallback: GpsStatus.Listener

    companion object {

        private const val PACKAGE_NAME = "com.sbh.background_locations"

        internal const val ACTION_FOREGROUND_ONLY_LOCATION_BROADCAST =
                "$PACKAGE_NAME.action.FOREGROUND_ONLY_LOCATION_BROADCAST"

        internal const val EXTRA_LOCATION = "$PACKAGE_NAME.extra.LOCATION"

        private const val EXTRA_CANCEL_LOCATION_TRACKING_FROM_NOTIFICATION =
                "$PACKAGE_NAME.extra.CANCEL_LOCATION_TRACKING_FROM_NOTIFICATION"

        var mCurrentLocation: Location? = null
        var m_serviceStatus = ServiceStatus.SERVICE_STOPPED

        enum class ServiceStatus constructor(value: Int) {
            PERMISSION_DENIED(0),
            SERVICE_STOPPED(1),
            SERVICE_STARTED(2),
            HAS_LOCATION(3),
            SERVICE_PAUSED(4);

            var value: Int = 0
                internal set

            init {
                this.value = value
            }
        }

        private var m_geoHashRTFilter: GeohashRTFilter? = null
        fun getGeoHashRTFilter(): GeohashRTFilter? {
            return m_geoHashRTFilter
        }

        class Settings(internal val accelerationDeviation: Double,
                       internal val gpsMinDistance: Int,
                       internal val gpsMinTime: Int,
                       internal val positionMinTime: Int,
                       private val geoHashPrecision: Int,
                       private val geoHashMinPointCount: Int,
                       internal val sensorFrequencyHz: Double,
//                       private val logger: ILogger,
                       internal val filterMockGpsCoordinates: Boolean,
                       internal val mVelFactor: Double,
                       internal val mPosFactor: Double)

        var defaultSettings = Settings(
                Utils.ACCELEROMETER_DEFAULT_DEVIATION,
                Utils.GPS_MIN_DISTANCE,
                Utils.GPS_MIN_TIME,
                Utils.SENSOR_POSITION_MIN_TIME,
                Utils.GEOHASH_DEFAULT_PREC,
                Utils.GEOHASH_DEFAULT_MIN_POINT_COUNT,
                Utils.SENSOR_DEFAULT_FREQ_HZ,
//                null,
                true,
                Utils.DEFAULT_VEL_FACTOR,
                Utils.DEFAULT_POS_FACTOR
        )

        private var m_settings: Settings? = defaultSettings
        private var m_locationManager: LocationManager? = null
        private var m_powerManager: PowerManager? = null
        private var m_wakeLock: PowerManager.WakeLock? = null

        private var m_gpsEnabled = false
        private var m_sensorsEnabled = false

        private var m_activeSatellites = 0
        private var m_lastLocationAccuracy = 0f
        private var m_gnssStatus: GnssStatus? = null
        private var m_gpsStatus: GpsStatus? = null

        /**/
        private var m_kalmanFilter: GPSAccKalmanFilter? = null
        private var m_eventLoopTask: SensorDataEventLoopTask? = null
        private val m_lstSensors: MutableList<Sensor> = ArrayList()
        private var m_sensorManager: SensorManager? = null
        private var m_magneticDeclination = 0.0

        /*accelerometer + rotation vector*/
        private val sensorTypes = intArrayOf(Sensor.TYPE_LINEAR_ACCELERATION, Sensor.TYPE_ROTATION_VECTOR)

        private val rotationMatrix = FloatArray(16)
        private val rotationMatrixInv = FloatArray(16)
        private val absAcceleration = FloatArray(4)
        private val linearAcceleration = FloatArray(4)
        //!
        private val m_sensorDataQueue = PriorityBlockingQueue<SensorGpsDataItem>()
    }

    @SuppressLint("InvalidWakeLockTag")
    override fun onCreate() {
        Log.d(TAG, "onCreate()")

        notificationManager = getSystemService(Context.NOTIFICATION_SERVICE) as NotificationManager
        m_locationManager = getSystemService(Context.LOCATION_SERVICE) as LocationManager
        m_sensorManager = getSystemService(Context.SENSOR_SERVICE) as SensorManager
        m_powerManager = getSystemService(Context.POWER_SERVICE) as PowerManager
        m_wakeLock = m_powerManager!!.newWakeLock(PowerManager.PARTIAL_WAKE_LOCK, TAG)
        firebaseDatabaseService = FirebaseDatabaseService.getInstance(this)

        if (m_sensorManager == null) {
            m_sensorsEnabled = false
            return  //todo handle somehow
        }

        for (st in sensorTypes) {
            val sensor = m_sensorManager!!.getDefaultSensor(st)
            if (sensor == null) {
                Log.d(TAG, String.format("Couldn't get sensor %d", st))
                continue
            }
            m_lstSensors.add(sensor)
        }

        locationRequest = LocationRequest()
        locationRequest.interval = UPDATE_INTERVAL_IN_MILLISECONDS
        locationRequest.fastestInterval = FASTEST_UPDATE_INTERVAL_IN_MILLISECONDS
        locationRequest.priority = LocationRequest.PRIORITY_HIGH_ACCURACY
        locationRequest.smallestDisplacement = 2f
        locationRequest.maxWaitTime = MAX_WAIT_TIME

        fusedLocationProviderClient = LocationServices.getFusedLocationProviderClient(this)

        locationCallback = object : LocationCallback() {
            override fun onLocationResult(locationResult: LocationResult?) {
                super.onLocationResult(locationResult)

                if (locationResult != null) {
                    if (locationResult.lastLocation != null) {
                        onNewLocation(locationResult.lastLocation)
                    } else {
                        Log.d(TAG, "Location missing in callback.")
                    }
                }
            }
        }
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
            gnssStatusCallback = object : GnssStatus.Callback() {
                override fun onSatelliteStatusChanged(status: GnssStatus?) {
                    super.onSatelliteStatusChanged(status)
                    m_gnssStatus = status
                }
            }
        } else {
            val context = this
            gpsStatusCallback = GpsStatus.Listener {
                if (checkSelfPermission(context, Manifest.permission.ACCESS_FINE_LOCATION) == PackageManager.PERMISSION_GRANTED) {
                    m_gpsStatus = m_locationManager!!.getGpsStatus(m_gpsStatus)
                }

                var activeSatellites = 0
                if (m_gpsStatus != null) {
                    for (satellite in m_gpsStatus!!.satellites) {
                        activeSatellites += if (satellite.usedInFix()) 1 else 0
                    }

                    if (activeSatellites != 0) {
                        m_activeSatellites = activeSatellites
                        //TODO fire gps changed
                        //for (locationServiceStatusInterface in m_locationServiceStatusInterfaces) {
                        //    locationServiceStatusInterface.GPSStatusChanged(this.m_activeSatellites)
                        //}
                    }
                }
            }
        }
    }

    override fun onStartCommand(intent: Intent, flags: Int, startId: Int): Int {
        super.onStartCommand(intent, flags, startId)
        Log.d(TAG, "onStartCommand()")

        val cancelLocationTrackingFromNotification =
                intent.getBooleanExtra(EXTRA_CANCEL_LOCATION_TRACKING_FROM_NOTIFICATION, false)

        if (cancelLocationTrackingFromNotification) {
            stopTrackingLocation()
            stopSelf()
        }
        // Tells the system not to recreate the service after it's been killed.
//        return START_NOT_STICKY
        return START_STICKY
    }

    override fun onBind(intent: Intent): IBinder? {
        Log.d(TAG, "onBind()")

        // MainActivity (client) comes into foreground and binds to service, so the service can
        // become a background services.
        stopForeground(true)
        serviceRunningInForeground = false
        configurationChange = false
        return localBinder
    }

    override fun onRebind(intent: Intent) {
        Log.d(TAG, "onRebind()")

        // MainActivity (client) returns to the foreground and rebinds to service, so the service
        // can become a background services.
        stopForeground(true)
        serviceRunningInForeground = false
        configurationChange = false
        super.onRebind(intent)
    }

    override fun onUnbind(intent: Intent): Boolean {
        Log.d(TAG, "onUnbind()")

        // MainActivity (client) leaves foreground, so service needs to become a foreground service
        // to maintain the 'while-in-use' label.
        // NOTE: If this method is called due to a configuration change in MainActivity,
        // we do nothing.
        if (!configurationChange && SharedPreferenceUtil.getLocationTrackingPref(this)) {
            Log.d(TAG, "Start foreground service")
            val notification = generateNotification()
            startForeground(NOTIFICATION_ID, notification)
            serviceRunningInForeground = true
        }

        // Ensures onRebind() is called if MainActivity (client) rebinds.
        return true
    }

    override fun onDestroy() {
        Log.d(TAG, "onDestroy()")
    }

    override fun onConfigurationChanged(newConfig: Configuration) {
        super.onConfigurationChanged(newConfig)
        Log.d(TAG, "onConfigurationChanged")
        configurationChange = true
    }

    fun startTrackingLocation() {
        Log.d(TAG, "startTrackingLocation()")
        SharedPreferenceUtil.saveLocationTrackingPref(this, true)

        m_wakeLock?.acquire(10 * 60 * 1000L /*10 minutes*/)
        m_sensorDataQueue.clear()

        if (checkSelfPermission(this, Manifest.permission.ACCESS_FINE_LOCATION) != PackageManager.PERMISSION_GRANTED) {
            m_serviceStatus = ServiceStatus.PERMISSION_DENIED
        } else {
            m_serviceStatus = ServiceStatus.SERVICE_STARTED
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
                m_locationManager?.unregisterGnssStatusCallback(gnssStatusCallback)
                m_locationManager?.registerGnssStatusCallback(gnssStatusCallback)
            } else {
                m_locationManager?.removeGpsStatusListener(gpsStatusCallback)
                m_locationManager?.addGpsStatusListener(gpsStatusCallback)
//                m_locationManager.removeUpdates(this)
//                m_locationManager.requestLocationUpdates(LocationManager.GPS_PROVIDER,
//                        m_settings?.gpsMinTime?.toLong(), m_settings?.gpsMinDistance?.toFloat(), this)
            }
        }

        m_sensorsEnabled = true
        for (sensor in m_lstSensors) {
            m_sensorManager?.unregisterListener(this, sensor)
            m_sensorsEnabled = m_sensorsEnabled and m_sensorManager?.registerListener(this, sensor, Utils.hertz2periodUs(m_settings!!.sensorFrequencyHz))!! == false
        }
        m_gpsEnabled = checkGps()

//        for (ilss in m_locationServiceStatusInterfaces) {
//            ilss.serviceStatusChanged(m_serviceStatus)
//            ilss.GPSEnabledChanged(m_gpsEnabled)
//        }

        m_eventLoopTask = SensorDataEventLoopTask(applicationContext, m_settings!!.positionMinTime.toLong(), this)
        m_eventLoopTask!!.needTerminate = false
        m_eventLoopTask!!.executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR)

        // Binding to this service doesn't actually trigger onStartCommand(). That is needed to
        // ensure this Service can be promoted to a foreground service, i.e., the service needs to
        // be officially started (which we do here).
        startService(Intent(applicationContext, LocationUpdatesService::class.java))

        try {
            fusedLocationProviderClient.requestLocationUpdates(locationRequest, locationCallback, Looper.myLooper())
//            callback?.onSuccess()
        } catch (unlikely: SecurityException) {
            SharedPreferenceUtil.saveLocationTrackingPref(this, false)
            Log.e(TAG, "Lost location permissions. Couldn't remove updates. $unlikely")
//            callback?.onFailure("Lost location permissions. Couldn't remove updates")
        }
    }

    fun checkGps(): Boolean {
        return m_locationManager?.isProviderEnabled(LocationManager.GPS_PROVIDER) ?: false
    }

    fun stopTrackingLocation() {

        Log.d(TAG, "stopTrackingLocation()")

        try {
            stopForeground(true)
            val removeTask = fusedLocationProviderClient.removeLocationUpdates(locationCallback)
            removeTask.addOnCompleteListener { task ->
                if (task.isSuccessful) {
                    Log.d(TAG, "Location Callback removed.")
                    notificationManager.cancel(NOTIFICATION_ID)
                    stopSelf()
                    stopForeground(true)
                } else {
                    Log.d(TAG, "Failed to remove Location Callback.")
                }
            }

            SharedPreferenceUtil.saveLocationTrackingPref(this, false)

        } catch (unlikely: SecurityException) {
            SharedPreferenceUtil.saveLocationTrackingPref(this, true)
            Log.e(TAG, "Lost location permissions. Couldn't remove updates. $unlikely")
        }
    }

    fun getCurrentLocation(callback: com.sbh.background_locations.callbacks.SBHLocationCallback) {
        Log.d(TAG, "getMCurrentLocation")
        fusedLocationProviderClient.lastLocation.addOnCompleteListener {
            if (it.isSuccessful && it.result != null) {
                Log.d(TAG, "getMCurrentLocation -> success")
                callback.onLocation(SBHLocation(location = it.result!!))
            } else {
                Log.d(TAG, "getMCurrentLocation -> error")
                callback.onError(0)
            }
        }
    }

    private fun onNewLocation(location: Location) {
        Log.d(TAG, "onNewLocation(): $location")
//        if (m_settings?.filterMockGpsCoordinates == true && location.isFromMockProvider) return
//
//        val x: Double = location.longitude
//        val y: Double = location.latitude
//        val posDev: Double = location.accuracy.toDouble()
//        val course: Double = location.bearing.toDouble()
//        val speed: Double = location.speed.toDouble()
//        val timeStamp: Long = Utils.nano2milli(location.elapsedRealtimeNanos)
//        val xVel: Double = speed * cos(course)
//        val yVel: Double = speed * sin(course)
//        //WARNING!!! here should be speed accuracy, but location.hasSpeedAccuracy()
//        // and location.getSpeedAccuracyMetersPerSecond() requares API 26
//        val velErr = location.accuracy * 0.1
//
//        val logStr = String.format("%d%d GPS : pos lat=%f, lon=%f, alt=%f, hdop=%f, speed=%f, bearing=%f, sa=%f",
//                Utils.LogMessageType.GPS_DATA.ordinal,
//                timeStamp, location.latitude,
//                location.longitude, location.altitude, location.accuracy,
//                location.speed, location.bearing, velErr)
////        log2File(logStr)
//        Log.d(TAG, logStr)
//
//        val f = GeomagneticField(
//                location.latitude.toFloat(),
//                location.longitude.toFloat(),
//                location.altitude.toFloat(),
//                timeStamp)
//        m_magneticDeclination = f.declination.toDouble()
//
//        if (m_kalmanFilter == null) {
////            log2File("%d%d KalmanAlloc : lon=%f, lat=%f, speed=%f, course=%f, m_accDev=%f, posDev=%f",
////                    Utils.LogMessageType.KALMAN_ALLOC.ordinal,
////                    timeStamp, x, y, speed, course, m_settings?.accelerationDeviation, posDev)
//            m_kalmanFilter = GPSAccKalmanFilter(
//                    true, //todo move to settings
//                    Coordinates.longitudeToMeters(x),
//                    Coordinates.latitudeToMeters(y),
//                    xVel,
//                    yVel,
//                    m_settings?.accelerationDeviation ?: 0.0,
//                    posDev,
//                    timeStamp,
//                    m_settings?.mVelFactor ?: 0.0,
//                    m_settings?.mPosFactor ?: 0.0)
//            return
//        }
//
//        val sdi = SensorGpsDataItem(
//                timeStamp, location.latitude, location.longitude, location.altitude,
//                SensorGpsDataItem.NOT_INITIALIZED,
//                SensorGpsDataItem.NOT_INITIALIZED,
//                SensorGpsDataItem.NOT_INITIALIZED,
//                location.speed,
//                location.bearing,
//                location.accuracy,
//                velErr,
//                m_magneticDeclination)
//        m_sensorDataQueue.add(sdi)

        // Existing Method
        mCurrentLocation = location

        // Notify anyone listening for broadcasts about the new location.
        val intent = Intent(ACTION_FOREGROUND_ONLY_LOCATION_BROADCAST)
        intent.putExtra(EXTRA_LOCATION, location)
        LocalBroadcastManager.getInstance(applicationContext).sendBroadcast(intent)

        // Update notification content if running as a foreground service.
        if (serviceRunningInForeground) {
            notificationManager.notify(NOTIFICATION_ID, generateNotification())
        }

        firebaseDatabaseService.updateLocation(location)
    }

    /*
     * Generates a BIG_TEXT_STYLE Notification that represent latest location.
     */
    private fun generateNotification(): Notification {
        Log.d(TAG, "generateNotification()")

        // Main steps for building a BIG_TEXT_STYLE notification:
        //      0. Get data
        //      1. Create Notification Channel for O+
        //      2. Build the BIG_TEXT_STYLE
        //      3. Set up Intent / Pending Intent for notification
        //      4. Build and issue the notification

        // 0. Get data
        val mainNotificationText = mCurrentLocation.toString()
//        val titleText = getString(R.string.app_name) TODO
        val titleText = "Background Location Service is Running"

        // 1. Create Notification Channel for O+ and beyond devices (26+).
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {

            val notificationChannel = NotificationChannel(
                    NOTIFICATION_CHANNEL_ID, titleText, NotificationManager.IMPORTANCE_LOW)

            // Adds NotificationChannel to system. Attempting to create an
            // existing notification channel with its original values performs
            // no operation, so it's safe to perform the below sequence.
            notificationManager.createNotificationChannel(notificationChannel)
        }

        // 2. Build the BIG_TEXT_STYLE.
        val bigTextStyle = NotificationCompat.BigTextStyle()
                .bigText(mainNotificationText)
                .setBigContentTitle(titleText)

        // 3. Set up main Intent/Pending Intents for notification.
//        val launchActivityIntent = Intent(this, MainActivity::class.java)

        val cancelIntent = Intent(this, LocationUpdatesService::class.java)
        cancelIntent.putExtra(EXTRA_CANCEL_LOCATION_TRACKING_FROM_NOTIFICATION, true)

        val servicePendingIntent = PendingIntent.getService(
                this, 0, cancelIntent, PendingIntent.FLAG_UPDATE_CURRENT)

//        val activityPendingIntent = PendingIntent.getActivity(
//                this, 0, launchActivityIntent, 0)

        // 4. Build and issue the notification.
        // Notification Channel Id is ignored for Android pre O (26).
        val notificationCompatBuilder =
                NotificationCompat.Builder(applicationContext, NOTIFICATION_CHANNEL_ID)

        return notificationCompatBuilder
//                .setStyle(bigTextStyle)
                .setContentTitle(titleText)
//                .setContentText(mainNotificationText)
                .setSmallIcon(R.mipmap.ic_service_launcher)
                .setOngoing(true)
                .setVisibility(NotificationCompat.VISIBILITY_PUBLIC)
//                .addAction(
//                        R.drawable.ic_launch, getString(R.string.launch_activity),
//                        activityPendingIntent
//                )
                .addAction(
                        android.R.drawable.ic_menu_close_clear_cancel,
                        "Disable Foreground Service",
                        servicePendingIntent
                )
                .build()
    }

    /**
     * Class used for the client Binder.  Since this service runs in the same process as its
     * clients, we don't need to deal with IPC.
     */
    inner class LocalBinder : Binder() {
        internal val service: LocationUpdatesService
            get() = this@LocationUpdatesService
    }

    class SensorDataEventLoopTask(val context: Context) : AsyncTask<Any, Any, Any>() {
        var needTerminate = false
        var deltaTMs: Long = 0
        var sdi: SensorGpsDataItem? = null
        lateinit var owner: LocationUpdatesService

        private fun handlePredict(sdi: SensorGpsDataItem) {
//            log2File("%d%d KalmanPredict : accX=%f, accY=%f",
//                    Utils.LogMessageType.KALMAN_PREDICT.ordinal,
//                    sdi.getTimestamp() as Long,
//                    sdi.getAbsEastAcc(),
//                    sdi.getAbsNorthAcc())
            m_kalmanFilter?.predict(sdi.getTimestamp(), sdi.getAbsEastAcc().toDouble(), sdi.getAbsNorthAcc().toDouble())
        }

        private fun handleUpdate(sdi: SensorGpsDataItem) {
            val xVel = sdi.getSpeed() * Math.cos(sdi.getCourse().toDouble())
            val yVel = sdi.getSpeed() * Math.sin(sdi.getCourse().toDouble())
//            log2File("%d%d KalmanUpdate : pos lon=%f, lat=%f, xVel=%f, yVel=%f, posErr=%f, velErr=%f",
//                    Utils.LogMessageType.KALMAN_UPDATE.ordinal,
//                    sdi.getTimestamp() as Long,
//                    sdi.getGpsLon(),
//                    sdi.getGpsLat(),
//                    xVel,
//                    yVel,
//                    sdi.getPosErr(),
//                    sdi.getVelErr()
//            )

            m_kalmanFilter?.update(
                    sdi.getTimestamp(),
                    Coordinates.longitudeToMeters(sdi.getGpsLon()),
                    Coordinates.latitudeToMeters(sdi.getGpsLat()),
                    xVel,
                    yVel,
                    sdi.getPosErr().toDouble(),
                    sdi.getVelErr()
            )
        }

        private fun locationAfterUpdateStep(sdi: SensorGpsDataItem): Location {
            val loc = Location(TAG)
            val pp = Coordinates.metersToGeoPoint(m_kalmanFilter!!.getCurrentX(),
                    m_kalmanFilter!!.getCurrentY())
            loc.latitude = pp.latitude
            loc.longitude = pp.longitude
            loc.altitude = sdi.getGpsAlt()
            val xVel: Double = m_kalmanFilter!!.getCurrentXVel()
            val yVel: Double = m_kalmanFilter!!.getCurrentYVel()
            val speed = sqrt(xVel * xVel + yVel * yVel) //scalar speed without bearing
            loc.bearing = sdi.getCourse()
            loc.speed = speed.toFloat()
            loc.time = System.currentTimeMillis()
            loc.elapsedRealtimeNanos = System.nanoTime()
            loc.accuracy = sdi.getPosErr()

            m_geoHashRTFilter?.filter(loc)

            return loc
        }

        constructor(context: Context, deltaTMs: Long, owner: LocationUpdatesService) : this(context) {
            this.deltaTMs = deltaTMs
            this.owner = owner
        }

        override fun doInBackground(vararg params: Any?): Any? {
            while (!needTerminate) {
                try {
                    Thread.sleep(deltaTMs)
                } catch (e: InterruptedException) {
                    e.printStackTrace()
                    continue //bad
                }

                var lastTimeStamp: Long = 0
                while (nextPoll() != null) {
                    if (sdi == null)
                        continue

                    if (sdi!!.getTimestamp() < lastTimeStamp) {
                        continue
                    }
                    lastTimeStamp = sdi!!.getTimestamp()

                    //warning!!!
                    if (sdi!!.getGpsLat().equals(SensorGpsDataItem.NOT_INITIALIZED)) {
                        handlePredict(sdi!!)
                    } else {
                        handleUpdate(sdi!!)
                        val loc = locationAfterUpdateStep(sdi!!)
                        publishProgress(loc)
                    }
                }
            }
            return null
        }

        private fun nextPoll(): SensorGpsDataItem? {
            sdi = m_sensorDataQueue.poll()
            return sdi
        }

        override fun onProgressUpdate(vararg values: Any?) {
            onLocationChangedImp(values[0] as Location)
        }

        private fun onLocationChangedImp(location: Location?) {
            Log.d(TAG, "onLocationChangedImp")
            if (location == null || location.latitude == 0.0 || location.longitude == 0.0 || location.provider == TAG)
                return

            m_serviceStatus = ServiceStatus.HAS_LOCATION
            mCurrentLocation = location
            m_lastLocationAccuracy = location.accuracy

            if (checkSelfPermission(owner, Manifest.permission.ACCESS_FINE_LOCATION) == PackageManager.PERMISSION_GRANTED) {
                if (Build.VERSION.SDK_INT < Build.VERSION_CODES.N) {
                    m_gpsStatus = m_locationManager?.getGpsStatus(m_gpsStatus)
                }
            }

            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
                if (m_gnssStatus != null)
                    m_activeSatellites = m_gnssStatus!!.satelliteCount
            } else {
                var activeSatellites = 0
                if (m_gpsStatus != null) {
                    for (satellite in m_gpsStatus!!.satellites) {
                        activeSatellites += if (satellite.usedInFix()) 1 else 0
                    }
                    m_activeSatellites = activeSatellites
                }
            }

            Log.d(TAG, "onLocationChangedImp => GOOD LOCATION")
            // Notify anyone listening for broadcasts about the new location.
            val intent = Intent(ACTION_FOREGROUND_ONLY_LOCATION_BROADCAST)
            intent.putExtra(EXTRA_LOCATION, location)
            LocalBroadcastManager.getInstance(context).sendBroadcast(intent)

            // TODO Update notification content if running as a foreground service.
//            if (serviceRunningInForeground) {
//                notificationManager.notify(NOTIFICATION_ID, generateNotification())
//            }

//            for (locationServiceInterface in m_locationServiceInterfaces) {
//                locationServiceInterface.locationChanged(location)
//            }
//            for (locationServiceStatusInterface in m_locationServiceStatusInterfaces) {
//                locationServiceStatusInterface.serviceStatusChanged(m_serviceStatus)
//                locationServiceStatusInterface.lastLocationAccuracyChanged(m_lastLocationAccuracy)
//                locationServiceStatusInterface.GPSStatusChanged(m_activeSatellites)
//            }
        }
    }
}