package com.sbh.background_locations.utils

/*
 * Copyright 2019 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import android.content.Context
import android.location.Location
import com.sbh.background_locations.R
import org.json.JSONObject

/**
 * Returns the `location` object as a human readable string.
 */
fun Location?.toText(): String {
    return if (this != null) {
        "($latitude, $longitude)"
    } else {
        "Unknown location"
    }
}

/**
 * Provides access to SharedPreferences for location to Activities and Services.
 */
internal object SharedPreferenceUtil {

    const val KEY_FOREGROUND_ONLY_ENABLED = "tracking_foreground_only_location"
    const val KEY_CONFIG = "config"

    /**
     * Returns true if requesting location updates, otherwise returns false.
     *
     * @param context The [Context].
     */
    fun getLocationTrackingPref(context: Context): Boolean =
            context.getSharedPreferences(
                    context.getString(R.string.preference_file_key), Context.MODE_PRIVATE)
                    .getBoolean(KEY_FOREGROUND_ONLY_ENABLED, false)

    /**
     * Stores the location updates state in SharedPreferences.
     * @param requestingLocationUpdates The location updates state.
     */
    fun saveLocationTrackingPref(context: Context, requestingLocationUpdates: Boolean) =
            context.getSharedPreferences(
                    context.getString(R.string.preference_file_key),
                    Context.MODE_PRIVATE)
                    .edit()
                    .putBoolean(KEY_FOREGROUND_ONLY_ENABLED, requestingLocationUpdates)
                    .apply()

    fun getConfig(context: Context): String? = context.getSharedPreferences(
            context.getString(R.string.preference_config_key), Context.MODE_PRIVATE)
            .getString(KEY_CONFIG, null)

    fun setConfig(context: Context, json: JSONObject?) = context.getSharedPreferences(
            context.getString(R.string.preference_config_key), Context.MODE_PRIVATE)
            .edit()
            .putString(KEY_CONFIG, json.toString())
            .apply()

    fun resetConfig(context: Context) = context.getSharedPreferences(
            context.getString(R.string.preference_config_key), Context.MODE_PRIVATE)
            .edit().clear().commit()
}